from gioconda._data import *
from gioconda._methods import *


class matrix_class():
    def __init__(self):
        self._create_data()

    def _create_data(self):
        self._data = []
        self._update_size()

    def _add_data(self, data, name = None):
        index = self._cols
        name = name if name is not None else index
        data = data_class(data, name, index)
        self._data = np.append(self._data, data)
        self._update_size()
        return data

    def _update_size(self):
        self._set_cols(len(self._data))
        self.set_rows(self.column(0)._rows if self._cols > 0 else 0)

    def set_rows(self, rows):
        self._rows = rows
        self._Rows = np.arange(self._rows)

    def _set_cols(self, cols):
        self._cols = cols
        self._Cols = np.arange(self._cols)

    def _update_indexes(self):
        [self._data[col]._set_index(col) for col in self._Cols]

    def _add_matrix(self, matrix, header = False):
        cols = len(matrix[0])
        columns = matrix[0] if header else [None] * cols
        matrix = matrix[1:] if header else matrix
        data = np.transpose(matrix)
        [self._add_data(data[i], columns[i]) for i in range(cols)]

    def set_columns(self, columns):
        [self.column(col)._set_name(columns[col]) for col in self._Cols]
        

    def column(self, col):
        col = self._index(col)
        return self._data[col]

    def _index(self, col):
        return self.columns().index(col) if isinstance(col, str) else col

    def _indexes(self, cols):
        return vectorize(self._index, cols)

    def _name(self, col):
        return self.column(col)._name

    def rename(self, col, name):
        self.column(col)._set_name(name)

    def columns(self, cols = None, index = False):
        cols = self._correct_cols(cols)
        return [self._name(col) for col in cols]

    def _correct_cols(self, cols):
        return np.array([col for col in cols if col in self._Cols or col in self.columns()]) if isinstance(cols, list) else self._Cols if cols is None else cols

    def _type(self, col):
        return self.column(col)._type

    def _types(self, cols = None):
        cols = self._correct_cols(cols)
        return [self.column(col)._type for col in cols]

    
    def get_section(self, rows = None, cols = None, nan = True, string = False, index = False):
        rows = self._correct_rows(rows)
        cols = self._correct_cols(cols)
        data = np.transpose([self.column(col).get_section(rows, nan = nan, string = string) for col in cols])
        data = np.concatenate([np.transpose([self._Rows]), data], axis = 1) if index else data
        return data

    def _correct_rows(self, rows):
        return self.column(0)._correct_rows(rows)


    def count(self, col, el, norm = False):
        return self.column(col).count(el, norm = norm)
    
    def counts(self, col, norm = False, nan = True):
        return self.column(col).counts(norm, nan)

    def unique(self, col, nan = True):
        return self.column(col).unique(nan)

    def distinct(self, col, nan = True):
        return self.column(col).distinct(nan)

    def _cross_count(self, col1, col2, val1, val2, norm = False):
        # return self.equal(col1, val1).count(col2, val2, norm = norm)
        rows1 = self.column(col1).equal(val1)
        count = np.count_nonzero(rows1 & self.column(col2).equal(val2))
        return 100 * count / np.count_nonzero(rows1) if norm else count


    def to_numerical(self, col, dictionary = None):
        return self.column(col).to_numerical(dictionary)
    
    def to_categorical(self, col):
        return self.column(col).to_categorical()
    
    def to_datetime(self, col, form = '%d/%m/%Y', delta_form = 'years'):
        return self.column(col).to_datetime(form, delta_form)
    

    def is_categorical(self, col):
        return self.column(col).is_categorical()

    def is_non_categorical(self, col):
        return self.column(col).is_non_categorical()

    def is_numerical(self, col):
        return self.column(col).is_numerical()
    
    def is_datetime(self, col):
        return self.column(col).is_datetime()

    def is_countable(self, col):
        return self.column(col).is_countable()

    def is_uncountable(self, col):
        return self.column(col).is_uncountable()
    
    def categorical_cols(self):
        return [self._name(col) for col in self._Cols if self.is_categorical(col)]
    
    def numerical_cols(self):
        return [self._name(col) for col in self._Cols if self.is_numerical(col)]
    
    def countable_cols(self):
        return [self._name(col) for col in self._Cols if self.is_countable(col)]


    def strip(self, cols = None):
        cols = self._correct_cols(cols)
        [self.column(col).strip() for col in cols]

    def replace(self, old, new, cols = None):
        cols = self._correct_cols(cols)
        [self.column(col).replace(old, new) for col in cols]


    def numerical_info(self):
        cols = self.countable_cols()
        infos = [self.column(col).numerical_info() for col in cols]
        header = list(infos[0].keys())
        table = [list(el.values()) for el in infos]
        table = [header] + table
        header = [''] + cols 
        table = tabulate(np.transpose(table), header = header)
        print(table + nl)

    def categorical_info(self, norm = False, cols = None, length = 10):
        cols = self._correct_cols(cols)
        cols = [col for col in cols if self.is_categorical(col)]
        [self.column(col)._print_counts(norm = norm, length = length) for col in cols]

    def _categorical_cross_counts(self, col1, col2, norm = False, length = 10):
        unique1 = list(self.unique(col1))[:length]; unique2 = list(self.unique(col2))[:length]
        counts = [[self._cross_count(col1, col2, u1, u2, norm = norm) for u2 in unique2] for u1 in unique1]
        table = [[unique1[i]] + counts[i] for i in range(len(counts))]
        header = [self._name(col1) + ' / ' + self._name(col2)] + unique2
        table = tabulate(table, header = header, decimals = 1)
        print(table)

    def _mixed_cross_counts(self, col1, col2, metric = 'mean', length = 10):
        unique = list(self.unique(col1))[ : length];
        value = list(self._get_encoding(col1, col2, metric, 1).values()) [ : length]
        name = self._name(col1) + ' / ' + self._name(col2)
        header = ['unique', metric]
        table = np.transpose([unique, value])
        table = tabulate(table, header = header, decimals = 1)
        print(table)

    def tab(self, col1, col2, norm = False, metric = 'mean', length = 30):
        if self.is_non_categorical(col1) and self.is_non_categorical(col2):
            print('Warning: At least one column should be categorical')
        elif self.is_categorical(col1) and self.is_categorical(col2):
            return self._categorical_cross_counts(col1, col2, norm = norm, length = length)
        elif self.is_categorical(col1) and self.is_non_categorical(col2):
            return self._mixed_cross_counts(col1, col2, metric, length)
        else:
            print('Warning: Columns switched')
            return self._mixed_cross_counts(col2, col1, metric, length)
        
    def plot(self, col, bins = 100):
        self.column(col).plot(bins)

    def cross_plot(self, col1, col2):
        plt.figure(0, figsize = (15, 8)); plt.clf()
        plt.scatter(self.column(col1).get_section(nan = True), self.column(col2).get_section(nan = True))
        plt.xlabel(self._name(col1)); plt.ylabel(self._name(col2))
        plt.xticks(rotation = 90) if self.is_categorical(col1) else None
        plt.tight_layout(); plt.pause(0.1); plt.show(block = 1); plt.clf(); plt.close()


    def _tabulate_data(self, rows = None, cols = None, header = True, index = False, decimals = 1):
        header = self.columns(cols, index) if header else None
        return tabulate(self.get_section(rows, cols, index = index, string = 1), header = header, decimals = decimals)

    def _tabulate_dimensions(self):
        return tabulate([[self._rows, self._cols]], header = ['rows', 'cols'])

    def _tabulate_types(self, cols = None):
        cols = self._Cols if cols is None else cols
        table = np.transpose([self._indexes(cols), self.columns(cols), self._types(cols)])
        return tabulate(table, header = ['i', 'column', 'type'])

    def _tabulate_info(self, cols = None):
        return self._tabulate_dimensions() + 2 * nl + self._tabulate_types(cols)

    def print(self, rows = None, cols = None, header = True, index = False, decimals = 1):
        rows = min(self._rows, th() - 8) if rows is None else rows
        print(self._tabulate_data(np.arange(rows), cols, header, index, decimals))
        print(nl + self._tabulate_dimensions())

    def __repr__(self):
        return self._tabulate_info()

    
    def __getitem__(self, cols):
        cols = [cols] if isinstance(cols, str) or isinstance(cols, int) else cols 
        new = matrix_class()
        for col in cols:
            new._data.append(self.column(col))
        new._update_size()
        return new


    def equal(self, col, value):
        return self.subset(self.column(col).equal(value))

    def not_equal(self, col, value):
        return self.subset(self.column(col).not_equal(value))
    
    def greater(self, col, value, equal = True):
        return self.subset(self.column(col).greater(value, equal))

    def lower(self, col, value, equal = True):
        return self.subset(self.column(col).lower(value, equal))

    def subset(self, rows = None):
        rows = self._correct_rows(rows)
        new = matrix_class()
        new._data = [data.subset(rows) for data in self._data]
        new._update_size()
        return new

    def select(self, cols = None):
        cols = self._correct_cols(cols)
        return self[cols]
    
    def part(self, start = None, end = None):
        start = 0 if start is None else max(0, start)
        end = self._rows if end is None else min(end, self._rows)
        return self.subset(np.arange(start, end))

    def copy(self):
        return copy(self)

    
    def simulate_categorical(self, name = None, categories = 5, nan_ratio = 0.1):
        categories = [random_word(5) for i in range(categories)] if isinstance(categories, int) else categories
        data = ['nan' if random.uniform(0, 1) < nan_ratio else random.choice(categories) for el in self._Rows]
        self._add_data(data, name)

    def simulate_numerical(self, name = None, mean = 0, std = 1, nan_ratio = 0.1):
        data = ['nan' if random.uniform(0, 1) < nan_ratio else random.normalvariate(mean, std) for el in self._Rows]
        self._add_data(data, name)
        self.to_numerical(name)

    def simulate_datetime(self, name = None, mean = "12/10/2000", std = 1, form = '%d/%m/%Y', delta_form = 'years', nan_ratio = 0.1):
        data = ['nan' if random.uniform(0, 1) < nan_ratio else random_datetime64(mean, std, form, delta_form) for el in self._Rows]
        self._add_data(data, name)
        self.to_datetime(name, form)

    def duplicate(self, col, name):
        data = self.column(col)._data
        self._add_data(data, name)
        self.column(name)._set_type(self.column(col)._type)
        return self

    def delete(self, col):
        index = self._index(col)
        self._data = np.delete(self._data, index)
        self._update_size()
        self._update_indexes()
        return self

    def sort_values(self, by):
        cols = [by] if isinstance(by, str) or isinstance(by, int) else by
        for col in cols:
            rows = self.column(col).argsort()
            self.column(col).sort(rows)
        return self

    def to_pandas(self):
        import pandas as pd
        df = pd.DataFrame(self.get_section(), columns = self.columns())
        return df

    def _get_encoding(self, col1, col2, metric = 'mean', string = False):
        unique = list(self.unique(col1)); l = len(unique)
        group = [self.equal(col1, u).column(col2) for u in unique]
        get_metric = lambda el: el.len() if metric == 'length' else el.sum() if metric == 'sum' else el.std() if metric == 'std' else el.mean() if metric == 'mean' else el
        values = [get_metric(el) for el in group]
        to_string = self.column(col1)._to_string
        values = list(map(to_string, values)) if string else values
        dictionary =  {unique[i]: values[i] for i in range(l)}
        return dictionary

    def encode(self, col1, col2, metric = 'mean', string = False):
        d = self._get_encoding(col1, col2, metric, string)
        self.to_numerical(col1, d)
        return self

    def to_pickle(self, path, log = True):
        import pickle
        print("writing pickle") if log else None
        with open(path, 'wb') as f:
            pickle.dump(self, f)
        print("pickle written!\n") if log else None

        
        



from functools import lru_cache as mem
from copy import deepcopy as copy
from math import isnan
import random, string
import datetime as dt
import numpy as np
import shutil
#from matplotlib import pyplot as plt

# System
tw = lambda: shutil.get_terminal_size()[0]
th = lambda: shutil.get_terminal_size()[1]

# Nan
nan = np.nan
nat = np.datetime64('NaT')

is_nan = lambda el: el is None or (isinstance(el, str) and el == 'nan') or (is_number(el) and isnan(el)) or (isinstance(el, np.datetime64) and np.isnan(el)) or (isinstance(el, np.timedelta64) and np.isnan(el))
is_number = lambda el: isinstance(el, float) or isinstance(el, int)
are_nan = lambda data: np.array([is_nan(el) for el in data], dtype = np.bool_)
are_not_nan = lambda data: np.array([not is_nan(el) for el in data], dtype = np.bool_)

# Data
#transpose = lambda data: list(map(list, zip(*data)))
vectorize = lambda method, data: np.vectorize(method)(data) if len(data) > 0 else np.array([])

# String
sp = ' '
vline = '│'
nl = '\n'
delimiter = sp * 2 + 1 * vline + sp * 0

bold = lambda string: '\x1b[1m' + string + '\x1b[0m'
pad = lambda string, length: string + sp * (length - len(string))

def tabulate(data, header = None, decimals = 1):
    cols = len(data[0]) if len(data) > 0 else 0; rows = len(data); Cols = range(cols)
    to_string = lambda el: str(round(el, decimals)) if is_number(el) else str(el)
    data = vectorize(to_string, data)
    data = np.concatenate([[header], data], axis = 0) if header is not None and len(data) > 0 else data if len(data) > 0 else [header]
    dataT = np.transpose(data)
    prepend_delimiter = lambda el: delimiter + el
    dataT = [vectorize(prepend_delimiter, dataT[i]) if i != 0 else dataT[i] for i in Cols]
    lengths = [max(vectorize(len, data)) for data in dataT]
    Cols = np.array(Cols)[np.cumsum(lengths) <= tw()]
    dataT = [vectorize(lambda el: pad(el, lengths[i]), dataT[i]) for i in Cols]
    data = np.transpose(dataT)
    lines = [''.join(line) for line in data]
    if len(lines) > 0 and header is not None:
        lines[0] = bold(lines[0])
    out = nl.join(lines)
    return out

def random_word(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))



# Datetime
def string_to_datetime64(string, form):
    return np.datetime64(dt.datetime.strptime(string, form)) if string != 'nan' else nat

strings_to_datetime64 = lambda data, form: np.array([string_to_datetime64(el, form) for el in data], dtype = np.datetime64)

def mean_datetime64(dates):
    std = dates[0].item() + dt.timedelta(seconds = np.mean(dates_to_seconds(dates)))
    return np.datetime64(std)

def median_datetime64(dates):
    std = dates[0].item() + dt.timedelta(seconds = np.median(dates_to_seconds(dates)))
    return np.datetime64(std)

def std_datetime64(dates):
    std = dt.timedelta(seconds = np.std(dates_to_seconds(dates)))
    return np.timedelta64(std)

def dates_to_seconds(dates):
    dates = [(el - dates[0]).item().total_seconds() for el in dates]
    return dates

div = [1, 60, 60, 24, 30.44, 12]
div = list(map(float, np.cumprod(div)))
forms = ['seconds', 'minutes', 'hours', 'days', 'months', 'years']

time_to_string = lambda date, form: date.strftime(form)

def timedelta64_to_number(delta, delta_form):
    #delta = delta.item().timestamp()
    delta = delta.item().total_seconds()
    index = forms.index(delta_form)
    return delta / div[index]

def timedelta64_to_numbers(data, form):
    return vectorize(lambda delta: timedelta64_to_number(delta, form), data)
    

timedelta64_to_string = lambda delta, form: str(round(timedelta64_to_number(delta, form), 1))

def random_datetime64(mean, std, form, delta_form):
    mean = string_to_datetime64(mean, form).item().timestamp()
    index = forms.index(delta_form)
    std = std * div[index]
    res = random.normalvariate(mean, std)
    return dt.datetime.fromtimestamp(res).strftime(form)



# /usr/bin/env python3
import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    author = "Savino Piccolomo",
    author_email = "piccolomo@gmail.com",
    name = 'gioconda',
    version='1.6.0',
    description = 'handles tabular data',
    long_description = README,
    long_description_content_type = "text/markdown",  
    license = "MIT",
    url = 'https://github.com/piccolomo/gioconda',
    packages = find_packages(),
    python_requires = ">=3.10.0, <=3.11.4",
    include_package_data = True,
    install_requires = ["numpy >= 1.24.0", "matplotlib >= 3.7.0"],
    classifiers = []
    )



from gioconda._matrix import matrix_class
from gioconda._data import nl
import os, sys
import inspect 

# File Utilities
script_folder = lambda: os.path.abspath(os.path.join(inspect.getfile(sys._getframe(1)), os.pardir))
source_folder = os.path.dirname(os.path.realpath(__file__))
test_data_path = os.path.join(source_folder, 'test_data.csv')

join = os.path.join

def _read_lines(path, log = True):
    print("reading text lines in", path) if log else None
    with open(path, 'r', encoding = "utf-8") as file:
        text = file.readlines()
    text = [line for line in text if line != nl]
    print("text lines read!\n") if log else None
    return text

def _split_lines(lines, delimiter = ','):
    return [line.replace("\n", "").split(delimiter) for line in lines]

def read(file_name, delimiter = ',', header = False, log = True):
    lines = _read_lines(file_name, log = log)
    print('loading data') if log else None
    matrix = _split_lines(lines, delimiter)
    data = matrix_class()
    data._add_matrix(matrix, header)
    print('data loaded!\n') if log else None
    return data

def read_pickle(path, log = True):
    import pickle
    print("reading pickle", path) if log else None
    with open(path, 'rb') as f:
        data = pickle.load(f)
    print("pickle read!\n") if log else None
    return data



from gioconda._methods import *
from matplotlib import pyplot as plt


class data_class():
    def __init__(self, data = [], name = '', index = 'none'):
        self._set_data(data)
        self._set_name(name)
        self._set_type()
        self._set_index(index)
        self._set_forms()

    def _set_data(self, data = []):
        self._data = np.array(data)
        self._update_length()

    def _set_name(self, name = 'none'):
        self._name = name


    def _set_type(self, type = 'categorical'):
        self._type = type

    def _set_index(self, index = 'none'):
        self._index = index

    def _set_forms(self, form = '%d/%m/%Y', delta_form = 'years'):
        self._form = form
        self._delta_form = delta_form
        
    def _update_length(self):
        self._rows = len(self._data)
        self._Rows = np.arange(self._rows)


    def get(self, row, string = False):
        el = self._data[row]
        el = self._to_string(el) if string else el
        return el

    def get_section(self, rows = None, nan = True, string = False):
        rows = self._correct_rows(rows)
        data = self._data[rows]
        data = data if nan else data[are_not_nan(data)]
        data = self._to_strings(data) if string else data
        return data
    
    def _correct_rows(self, rows):
        return np.array([row for row in rows if row in self._Rows]) if isinstance(rows, list) else self._Rows if rows is None else rows
    

    @mem(maxsize = None)
    def counts(self, norm = False, nan = True):
        u, v = np.unique(self._data, return_counts = 1)
        nans = are_nan(u)
        u = u if nan else u[~nans]
        v = v if nan else v[~nans]
        t = sum(v)
        c = {u[i] : v[i] for i in range(len(u))}
        c = {key: (100 * count / t) for key, count in c.items()} if norm else c
        return c

    def count(self, value, norm = False):
        counts = self.counts(norm)
        return counts[value] if value in counts else 0
        #return 100 * count / self.rows if norm else count
    
    def count_nan(self, norm = False):
        c = np.count_nonzero(are_nan(self._data))
        return nan if self._rows == 0 else 100 * c / self._rows if norm else c

    @mem(maxsize = None)
    def unique(self, nan = True):
        return np.array(list(self.counts(False, nan).keys()))

    def distinct(self, nan = True):
        return len(self.unique(nan))

    def mode(self):
        unique = self.unique()
        return unique[0] if len(unique) > 0 else nan


    def to_categorical(self):
        self._data = self._to_strings(self._data)
        self._update_length()
        self._set_type('categorical')
        return self

    def to_numerical(self, dictionary = None):
        self._data = self._to_numbers(dictionary)
        self._update_length()
        self._set_type('numerical')
        return self

    def _to_numbers(self, dictionary = None):
        if self.is_datetime():
            return  np.array([timedelta64_to_number(el, self._delta_form) if not is_nan(el) else nan for el in self._data - self.min()])
        elif dictionary is None:
            return np.array([np.float64(el) if not is_nan(el) else nan for el in self._data])
        else:
            return np.array([dictionary[el] if not is_nan(el) else nan for el in self._data])


    def to_datetime(self, form = '%d/%m/%Y', delta_form = 'years'):
        self._data = strings_to_datetime64(self._data, form)
        self._update_length()
        self._set_type('datetime')
        self._set_forms(form, delta_form)
        return self


    def is_mixed(self):
        return self._type == 'mixed'

    def is_categorical(self):
        return self._type == 'categorical'

    def is_non_categorical(self):
        return not self.is_categorical()

    def is_numerical(self):
        return self._type == 'numerical'
    
    def is_datetime(self):
        return self._type == 'datetime'
    
    def is_countable(self):
        return self.is_numerical() or self.is_datetime()
    
    def is_uncountable(self):
        return not self.is_countable()
    
    
    
    def strip(self):
        self._apply(lambda string: string.strip()) if self.is_categorical() else print('not categorical')

    def replace(self, old, new):
        self._apply(lambda string: string.replace(old, new)) if self.is_categorical() else print('not categorical')

    def _apply(self, function):
        data = vectorize(function, self._data)
        self._set_data(data)

    def len(self):
        return self._rows

    def sum(self):
        data = self.get_section(nan = False); l = len(data)
        return nan if l == 0 or self.is_uncountable() else mean_datetime64(data) if self.is_datetime() else np.sum(data)
        
    def min(self, string = False):
        data = self.get_section(nan = False); l = len(data)
        m = nan if l == 0 or self.is_uncountable() else np.min(data)
        return m
    
    def max(self):
        data = self.get_section(nan = False); l = len(data)
        m = nan if l == 0 or self.is_uncountable() else np.max(data)
        return m

    def span(self):
        m, M = self.min(), self.max()
        s = M - m
        return s if not self.is_datetime() else s
    
    def std(self):
        data = self.get_section(nan = False); l = len(data)
        return nan if l == 0 or self.is_uncountable() else std_datetime64(data) if self.is_datetime() else np.std(data)
    
    def density(self):
        std = self.std().item().total_seconds() if self.is_datetime() else self.std()
        span = self.span().item().total_seconds() if self.is_datetime() else self.span()
        return 100 * std / span if span != 0 else np.inf

    def mean(self):
        data = self.get_section(nan = False); l = len(data)
        return nan if l == 0 or self.is_uncountable() else mean_datetime64(data) if self.is_datetime() else np.mean(data)

    def median(self):
        data = self.get_section(nan = False); l = len(data)
        return nan if l == 0 or self.is_uncountable() else median_datetime64(data) if self.is_datetime() else np.median(data)

    def _get_numerical_data(self):
        data = self.get_section(nan = False)
        data = [el.item().timestamp() for el in data] if self.is_datetime() else data
        return data if self.is_countable() else []
    
    def multiply(self, k):
        self._set_data(k * self._data) if self.is_numerical() else print('not numerical')
        

    def _tabulate_counts(self, norm = False, length = 10):
        header = [self._name, 'count']
        counts = list(self.counts(norm = norm).items())[: length]
        table = tabulate(counts, header = header) + nl
        return table

    def _print_counts(self, norm = False, length = 10):
        print(self._tabulate_counts(norm, length))
        
    @mem(maxsize = None)
    def _basic_info(self):
        return {'name': self._name, 'index': self._index, 'type': self._type, 'rows': self._rows, 'nan': self.count_nan(), 'unique': self.distinct(nan=False)}

    @mem(maxsize=None)
    def numerical_info(self):
        info = {'min': self.min(), 'max': self.max(), 'span': self.span(), 'nan': self.count_nan(1), 'mean': self.mean(), 'median': self.median(), 'mode': self.mode(), 'std': self.std(), 'density': self.density()}
        return {k : self._to_string(info[k]) for k in info.keys()}
    
    def info(self):
        info = self._basic_info()
        info.update(self.numerical_info()) if self.is_countable() else None
        return info

    def plot(self, bins = 100):
        plt.figure(0, figsize = (15, 8)); plt.clf()
        bins = min(bins, len(self.unique())) if self.is_countable() else None
        plt.hist(self.get_section(nan = False), bins = bins) if self.is_countable() else None
        plt.bar(self.counts().keys(), self.counts().values()) if self.is_uncountable() else None
        plt.xlabel(self._name); plt.ylabel('count')
        plt.xticks(rotation = 90) if self.is_categorical() else None
        plt.tight_layout(); plt.pause(0.1); plt.show(block = 1); plt.clf(); plt.close()
    
    
    def _tabulate_info(self):
        info = self.info()
        table = [list(info.keys()), list(info.values())]
        table = tabulate(np.transpose(table))
        return table + nl

    def print(self):
        print(self._tabulate_info())

    def _get_sample_data(self, length = 10):
        m = min(self._rows, length)
        start = [self._to_string(self._data[i]) for i in range(0, m)]
        end = [self._to_string(self._data[i]) for i in range(-m, 0)]
        out = ', '.join(start)
        out += ' ... ' if self._rows > length else ''
        out += ', '.join(end) if self._rows > 2 * length else ''
        return out

    def _to_string(self, el):
        return 'nan' if is_nan(el) else el.item().strftime(self._form) if isinstance(el, np.datetime64) else timedelta64_to_string(el, self._delta_form) if isinstance(el, np.timedelta64) else str(round(el, 2)) if is_number(el) else str(el)

    def _to_strings(self, data):
        return np.array([self._to_string(el) for el in data])

    def __str__(self):
        return self._tabulate_info() + nl + 0 * sp + self._get_sample_data()

    def __repr__(self):
        return str(self)

    def __getitem__(self, row):
        return self._data[row]


        

    def equal(self, value):
        data = self.get_section(string = 1) if isinstance(value, str) and self.is_datetime() else self._data
        return are_nan(self._data) if is_nan(value) else self._data == value

    def not_equal(self, value):
        return ~ self.equal(value)

    def higher(self, value, equal = False):
        return self._data >= value if equal else self._data > value

    def lower(self, value, equal = False):
        return ~self.higher(value, not equal)


    def copy(self):
        return copy(self)

    def empty(self):
        new = self.__class__([])
        new._set_name(self._name)
        new._set_type(self._type)
        new._set_index(self._index)
        return new

    def part(self, a = None, b = None):
        a = 0 if a is None else a
        b = self._rows if b is None else b
        return self.subset(np.arange(a, b))
    
    def subset(self, rows):
        new = self.empty()
        new._set_data(self.get_section(rows))
        return new

    def argsort(self):
        return np.argsort(self._data)

    def sort(self, rows):
        self._data = self._data[rows]

    def tolist(self):
        return self._data.tolist()
